// ****************************************************************************
// 
// FLV Extract
// Copyright (C) 2006  Moitah (moitah@yahoo.com)
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// ****************************************************************************

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace JDP {
	interface IAudioWriter {
		void WriteChunk(byte[] chunk);
		void Finish();
	}

	interface IVideoWriter {
		void WriteChunk(byte[] chunk, uint timeStamp, int frameType);
		void Finish(FractionUInt32 averageFrameRate);
	}

	class FLVFile {
		private string _path, _pathBase;
		private FileStream _fs;
		private long _fileOffset, _fileLength;
		private IAudioWriter _audioWriter;
		private IVideoWriter _videoWriter;
		private StreamWriter _timeCodeSW;
		private List<uint> _videoTimeStamps;
		private bool _extractAudio, _extractVideo, _extractTimeCodes;
		private FractionUInt32 _averageFrameRate, _trueFrameRate;

		public FLVFile(string path) {
			_path = path;
			_fs = new FileStream(path, FileMode.Open, FileAccess.Read, FileShare.Read, 65536);
			_fileOffset = 0;
			_fileLength = _fs.Length;
		}

		public void Close() {
			_fs.Close();
		}

		public FractionUInt32 AverageFrameRate {
			get {
				return _averageFrameRate;
			}
		}

		public FractionUInt32 TrueFrameRate {
			get {
				return _trueFrameRate;
			}
		}

		public void ExtractStreams(bool extractAudio, bool extractVideo, bool extractTimeCodes) {
			uint dataOffset, flags, prevTagSize;
			string ext;

			ext = Path.GetExtension(_path).ToLower();
			if ((ext == ".avi") || (ext == ".mp3") || (ext == ".txt")) {
				// Can't have the same extension as the files we output
				throw new Exception("Please change the extension of your FLV file.");
			}
			_pathBase = _path.Substring(0, _path.Length - ext.Length);

			_extractAudio = extractAudio;
			_extractVideo = extractVideo;
			_extractTimeCodes = extractTimeCodes;
			_videoTimeStamps = new List<uint>();

			Seek(0);
			if (ReadUInt32() != 0x464C5601) {
				throw new Exception("This isn't a FLV file.");
			}

			flags = ReadUInt8();
			dataOffset = ReadUInt32();

			Seek(dataOffset);

			prevTagSize = ReadUInt32();
			while (_fileOffset < _fileLength) {
				ReadTag();
				prevTagSize = ReadUInt32();
			}

			_averageFrameRate = CalculateAverageFrameRate();
			_trueFrameRate = CalculateTrueFrameRate();

			if (_audioWriter != null) {
				_audioWriter.Finish();
				_audioWriter = null;
			}
			if (_videoWriter != null) {
				_videoWriter.Finish(_averageFrameRate);
				_videoWriter = null;
			}
			if (_timeCodeSW != null) {
				_timeCodeSW.Close();
				_timeCodeSW = null;
			}
		}

		private void ReadTag() {
			uint tagType, dataSize, timeStamp, streamID, mediaInfo;
			byte[] data;

			// Read tag header
			tagType = ReadUInt8();
			dataSize = ReadUInt24();
			timeStamp = ReadUInt24();
			timeStamp |= ReadUInt8() << 24; 
			streamID = ReadUInt24();

			// Read tag data
			if (dataSize == 0) {
				return;
			}
			mediaInfo = ReadUInt8();
			dataSize -= 1;
			data = ReadBytes((int)dataSize);

			if (tagType == 0x8) {  // Audio
				if (_audioWriter == null) {
					_audioWriter = _extractAudio ? GetAudioWriter(mediaInfo) : new DummyAudioWriter();
				}
				_audioWriter.WriteChunk(data);
			}
			else if (tagType == 0x9) { // Video
				if (_videoWriter == null) {
					_videoWriter = _extractVideo ? GetVideoWriter(mediaInfo) : new DummyVideoWriter();
				}
				if (_extractTimeCodes) {
					if (_timeCodeSW == null) {
						_timeCodeSW = new StreamWriter(_pathBase + ".txt", false, Encoding.ASCII);
						_timeCodeSW.WriteLine("# timecode format v2");
					}
					_timeCodeSW.WriteLine(timeStamp.ToString());
				}
				_videoTimeStamps.Add(timeStamp);
				_videoWriter.WriteChunk(data, timeStamp, (int)((mediaInfo & 0xF0) >> 4));
			}
		}

		private IAudioWriter GetAudioWriter(uint mediaInfo) {
			uint format = (mediaInfo & 0xF0) >> 4;

			if (format == 0x2) { // MP3
				return new MP3Writer(_pathBase + ".mp3");
			}
			else {
				return new DummyAudioWriter();
			}
		}

		private IVideoWriter GetVideoWriter(uint mediaInfo) {
			uint codecID = mediaInfo & 0x0F;

			if ((codecID == 2) || (codecID == 4) || (codecID == 5)) {
				return new AVIWriter(_pathBase + ".avi", (int)codecID);
			}
			else {
				return new DummyVideoWriter();
			}
		}

		private FractionUInt32 CalculateAverageFrameRate() {
			FractionUInt32 frameRate;
			int frameCount = _videoTimeStamps.Count;

			if (frameCount > 1) {
				frameRate.N = (uint)(frameCount - 1) * 1000;
				frameRate.D = _videoTimeStamps[frameCount - 1] - _videoTimeStamps[0];
				frameRate.Reduce();
			}
			else {
				frameRate.N = 25;
				frameRate.D = 1;
			}

			return frameRate;
		}

		private FractionUInt32 CalculateTrueFrameRate() {
			FractionUInt32 frameRate;
			Dictionary<uint, uint> deltaCount = new Dictionary<uint, uint>();
			int i, threshold;
			uint delta, count, minDelta;

			// Calculate the distance between the timestamps, count how many times each delta appears
			for (i = 1; i < _videoTimeStamps.Count; i++) {
				int deltaS = (int)((long)_videoTimeStamps[i] - (long)_videoTimeStamps[i - 1]);

				if (deltaS < 0) continue;
				delta = (uint)deltaS;

				if (deltaCount.ContainsKey(delta)) {
					deltaCount[delta] += 1;
				}
				else {
					deltaCount.Add(delta, 1);
				}
			}

			threshold = _videoTimeStamps.Count / 10;
			minDelta = UInt32.MaxValue;

			// Find the smallest delta that made up at least 10% of the frames (grouping in delta+1
			// because of rounding, e.g. a NTSC video will have deltas of 33 and 34 ms)
			foreach (KeyValuePair<uint, uint> deltaItem in deltaCount) {
				delta = deltaItem.Key;
				count = deltaItem.Value;

				if (deltaCount.ContainsKey(delta + 1)) {
					count += deltaCount[delta + 1];
				}

				if ((count >= threshold) && (delta < minDelta)) {
					minDelta = delta;
				}
			}

			// Calculate the frame rate based on the smallest delta, and delta+1 if present
			if (minDelta != UInt32.MaxValue) {
				uint totalTime, totalFrames;
				
				count = deltaCount[minDelta];
				totalTime = minDelta * count;
				totalFrames = count;

				if (deltaCount.ContainsKey(minDelta + 1)) {
					count = deltaCount[minDelta + 1];
					totalTime += (minDelta + 1) * count;
					totalFrames += count;
				}

				if (totalTime != 0) {
					frameRate.N = totalFrames * 1000;
					frameRate.D = totalTime;
					frameRate.Reduce();
					return frameRate;
				}
			}

			// Default if the frame rate couldn't be calculated
			frameRate.N = 25;
			frameRate.D = 1;
			return frameRate;
		}

		private void Seek(long offset) {
			_fs.Seek(offset, SeekOrigin.Begin);
			_fileOffset = offset;
		}

		private uint ReadUInt8() {
			_fileOffset += 1;
			return (uint)_fs.ReadByte();
		}

		private uint ReadUInt24() {
			byte[] x = new byte[4];
			_fs.Read(x, 1, 3);
			_fileOffset += 3;
			return ByteOrder.Reverse(BitConverter.ToUInt32(x, 0));
		}

		private uint ReadUInt32() {
			byte[] x = new byte[4];
			_fs.Read(x, 0, 4);
			_fileOffset += 4;
			return ByteOrder.Reverse(BitConverter.ToUInt32(x, 0));
		}

		private byte[] ReadBytes(int length) {
			byte[] buff = new byte[length];
			_fs.Read(buff, 0, length);
			_fileOffset += length;
			return buff;
		}
	}

	class DummyAudioWriter : IAudioWriter {
		public DummyAudioWriter() {
		}

		public void WriteChunk(byte[] chunk) {
		}

		public void Finish() {
		}
	}

	class DummyVideoWriter : IVideoWriter {
		public DummyVideoWriter() {
		}

		public void WriteChunk(byte[] chunk, uint timeStamp, int frameType) {
		}

		public void Finish(FractionUInt32 averageFrameRate) {
		}
	}

	class MP3Writer : IAudioWriter {
		FileStream _fs;

		public MP3Writer(string path) {
			_fs = new FileStream(path, FileMode.Create, FileAccess.Write, FileShare.Read, 65536);
		}

		public void WriteChunk(byte[] chunk) {
			_fs.Write(chunk, 0, chunk.Length);
		}

		public void Finish() {
			_fs.Close();
		}
	}

	class AVIWriter : IVideoWriter {
		private BinaryWriter _bw;
		private int _codecID;
		private int _width, _height, _frameCount;
		private uint _moviDataSize, _indexChunkSize;
		private List<uint> _index;

		// Chunk:          Off:  Len:
		//
		// RIFF AVI          0    12
		//   LIST hdrl      12    12
		//     avih         24    64
		//     LIST strl    88    12
		//       strh      100    64
		//       strf      164    48
		//   LIST movi     212    12
		//     (frames)    224   ???
		//   idx1          ???   ???

		public AVIWriter(string path, int codecID) {
			if ((codecID != 2) && (codecID != 4) && (codecID != 5)) {
				throw new Exception("Unsupported video codec.");
			}
			_codecID = codecID;

			FileStream fs = new FileStream(path, FileMode.Create, FileAccess.Write, FileShare.Read);

			_bw = new BinaryWriter(fs);

			WriteFourCC("RIFF");
			_bw.Write((uint)0); // chunk size
			WriteFourCC("AVI ");

			WriteFourCC("LIST");
			_bw.Write((uint)192);
			WriteFourCC("hdrl");

			WriteFourCC("avih");
			_bw.Write((uint)56);
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0x10);
			_bw.Write((uint)0); // frame count
			_bw.Write((uint)0);
			_bw.Write((uint)1);
			_bw.Write((uint)0);
			_bw.Write((uint)0); // width
			_bw.Write((uint)0); // height
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0);

			WriteFourCC("LIST");
			_bw.Write((uint)116);
			WriteFourCC("strl");

			WriteFourCC("strh");
			_bw.Write((uint)56);
			WriteFourCC("vids");
			WriteFourCC(CodecFourCC);
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0); // frame rate denominator
			_bw.Write((uint)0); // frame rate numerator
			_bw.Write((uint)0);
			_bw.Write((uint)0); // frame count
			_bw.Write((uint)0);
			_bw.Write((int)-1);
			_bw.Write((uint)0);
			_bw.Write((ushort)0);
			_bw.Write((ushort)0);
			_bw.Write((ushort)0); // width
			_bw.Write((ushort)0); // height

			WriteFourCC("strf");
			_bw.Write((uint)40);
			_bw.Write((uint)40);
			_bw.Write((uint)0); // width
			_bw.Write((uint)0); // height
			_bw.Write((ushort)1);
			_bw.Write((ushort)24);
			WriteFourCC(CodecFourCC);
			_bw.Write((uint)0); // biSizeImage
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0);

			WriteFourCC("LIST");
			_bw.Write((uint)0); // chunk size
			WriteFourCC("movi");

			_index = new List<uint>();
		}

		public void WriteChunk(byte[] chunk, uint timeStamp, int frameType) {
			int offset, len;

			offset = 0;
			if (_codecID == 4) offset = 1;
			if (_codecID == 5) offset = 4;
			len = Math.Max(chunk.Length - offset, 0);

			_index.Add((frameType == 1) ? (uint)0x10 : (uint)0);
			_index.Add(_moviDataSize + 4);
			_index.Add((uint)len);

			if ((_width == 0) && (_height == 0)) {
				GetFrameSize(chunk);
			}

			WriteFourCC("00dc");
			_bw.Write(len);
			_bw.Write(chunk, offset, len);

			if ((len % 2) != 0) {
				_bw.Write((byte)0);
				len++;
			}
			_moviDataSize += (uint)len + 8;
			_frameCount++;
		}

		private void GetFrameSize(byte[] chunk) {
			if (_codecID == 2) {
				// Reference: flv_h263_decode_picture_header from libavcodec's h263.c

				if ((chunk[0] != 0) || (chunk[1] != 0)) {
					return;
				}

				ulong x = ByteOrder.Reverse(BitConverter.ToUInt64(chunk, 2));
				int format;

				if (ReadBits(ref x, 1) != 1) {
					return;
				}
				ReadBits(ref x, 5);
				ReadBits(ref x, 8);

				format = ReadBits(ref x, 3);
				switch (format) {
					case 0:
						_width = ReadBits(ref x, 8);
						_height = ReadBits(ref x, 8);
						break;
					case 1:
						_width = ReadBits(ref x, 16);
						_height = ReadBits(ref x, 16);
						break;
					case 2:
						_width = 352;
						_height = 288;
						break;
					case 3:
						_width = 176;
						_height = 144;
						break;
					case 4:
						_width = 128;
						_height = 96;
						break;
					case 5:
						_width = 320;
						_height = 240;
						break;
					case 6:
						_width = 160;
						_height = 120;
						break;
					default:
						return;
				}
			}
			else if ((_codecID == 4) || (_codecID == 5)) {
				// Reference: CFLVSplitterFilter::CreateOutputs from Media Player Classic's FLVSplitter.cpp

				ulong x = ByteOrder.Reverse(BitConverter.ToUInt64(chunk, (_codecID == 4) ? 1 : 4));

				if ((ReadBits(ref x, 16) & 0x80fe) != 0x0046) {
					return;
				}

				_height = ReadBits(ref x, 8) * 16;
				_width = ReadBits(ref x, 8) * 16;
			}
		}

		private string CodecFourCC {
			get {
				if (_codecID == 2) {
					return "FLV1";
				}
				if ((_codecID == 4) || (_codecID == 5)) {
					return "FLV4";
				}
				return "NULL";
			}
		}

		private int ReadBits(ref ulong x, int length) {
			int r = (int)(x >> (64 - length));
			x <<= length;
			return r;
		}

		private void WriteIndexChunk() {
			uint indexDataSize = (uint)_frameCount * 16;

			WriteFourCC("idx1");
			_bw.Write(indexDataSize);

			for (int i = 0; i < _frameCount; i++) {
				WriteFourCC("00dc");
				_bw.Write(_index[(i * 3) + 0]);
				_bw.Write(_index[(i * 3) + 1]);
				_bw.Write(_index[(i * 3) + 2]);
			}

			_indexChunkSize = indexDataSize + 8;
		}

		public void Finish(FractionUInt32 averageFrameRate) {
			WriteIndexChunk();

			_bw.BaseStream.Seek(4, SeekOrigin.Begin);
			_bw.Write((uint)(224 + _moviDataSize + _indexChunkSize - 8));

			_bw.BaseStream.Seek(24 + 8, SeekOrigin.Begin);
			_bw.Write((uint)0);
			_bw.BaseStream.Seek(12, SeekOrigin.Current);
			_bw.Write((uint)_frameCount);
			_bw.BaseStream.Seek(12, SeekOrigin.Current);
			_bw.Write((uint)_width);
			_bw.Write((uint)_height);

			_bw.BaseStream.Seek(100 + 28, SeekOrigin.Begin);
			_bw.Write((uint)averageFrameRate.D);
			_bw.Write((uint)averageFrameRate.N);
			_bw.BaseStream.Seek(4, SeekOrigin.Current);
			_bw.Write((uint)_frameCount);
			_bw.BaseStream.Seek(16, SeekOrigin.Current);
			_bw.Write((ushort)_width);
			_bw.Write((ushort)_height);

			_bw.BaseStream.Seek(164 + 12, SeekOrigin.Begin);
			_bw.Write((uint)_width);
			_bw.Write((uint)_height);
			_bw.BaseStream.Seek(8, SeekOrigin.Current);
			_bw.Write((uint)(_width * _height * 6));

			_bw.BaseStream.Seek(212 + 4, SeekOrigin.Begin);
			_bw.Write((uint)(_moviDataSize + 4));

			_bw.Close();
		}

		private void WriteFourCC(string fourCC) {
			byte[] bytes = System.Text.Encoding.ASCII.GetBytes(fourCC);
			if (bytes.Length != 4) {
				throw new Exception("Invalid FourCC length.");
			}
			_bw.Write(bytes);
		}
	}

	struct FractionUInt32 {
		public uint N;
		public uint D;

		public double ToDouble() {
			return (double)N / (double)D;
		}

		public void Reduce() {
			uint gcd = GCD(N, D);
			N /= gcd;
			D /= gcd;
		}

		private uint GCD(uint a, uint b) {
			uint r;

			while (b != 0) {
				r = a % b;
				a = b;
				b = r;
			}

			return a;
		}

		public override string ToString() {
			return ToDouble().ToString() + " (" + N.ToString() + "/" + D.ToString() + ")";
		}
	}
}